
/* See file "monitor.c" for copyright and other information. */

#include <stdio.h>
#include <math.h>	  /* for pow(), sin(), cos(), and similar functions */
#include "globals.h"


typedef struct c_point                        /* record for complex number */
        { double real,imag; }
        complex;
void fast_fourier_transform();				  /* function prototype */

											  /* Kalman filter constants */
#define TOTAL_STATE_VARS	2		/* state = position, velocity */
#define	DELTA_T				1.0		/* time between measurements */
#define	MEAS_NOISE			1.0		/* units are distance */
#define	MAX_ACCEL			0.05	/* units are per second squared */


		/*
		** The main arousal analysis function.  It takes in an sIBI, and the
		** desired sampling rate, window size, and KF smoothing factor.  It returns
		** a bunch of stuff from the analysis (see list, below).  Locally, it
		** maintains a buffer of the synchronous IBIs used for analysis.  It also
		** maintains a buffer of recent power spectrum results.
		**
		** Returns the number of sIBIs needed before arousal can be computed.
		**
		** The first time the function is called, sIBI=-1 must be passed.
		** This sets up the local buffers.  To release the memory, the function should
		** be called a final time with an sIBI=-2.
		*/

int  ArousalAnalysis(int		sIBI,						/* input synchronous IBI */
					 int		LocalSamplingRate,			/* sampling rate (units are Hz) */
					 int		LocalWindowSize,			/* sampling window (units are sIBI's) */
					 double		LocalKFSmooth,				/* KF smoothing factor (unitless) */
					 int		*TotalData,					/* input/output - total data in buffer */
															/* set to zero on first call */
					 double		Power[TOTAL_FREQS],			/* returned - current power spectrum */
					 double		*CurrentMaxPower,			/* returned - current max power bar in spectrum */
					 int		*CurrentMaxPowerCPM,		/* returned - cycles per minute of max power bar */
					 double		*MeanRSA,					/* returned - mean RSA over all time */
					 double		*StdDevRSA,					/* returned - stddev RSA over all time */
					 int		*TotalMinutes,				/* returned - total minutes for RSA statistics */
					 double		*Arousal)					/* returned - current normalized arousal */

{
static unsigned short int	*DataBuffer;
static int					TotalRSA;						/* local count of total times Arousal has been computed */
int							i,j,bin;
double						ZScore,SumRSA,LogPower,hist_bin_width;
				/* Kalman filter variables */
static double	phi[TOTAL_STATE_VARS][TOTAL_STATE_VARS];		/* state transition matrix */
static double	M[TOTAL_STATE_VARS];							/* measurement matrix */
static double	Xtt[TOTAL_STATE_VARS],Xt1t[TOTAL_STATE_VARS];	/* t,t state matrix and t+1,t state matrix */
static double	Qt[TOTAL_STATE_VARS][TOTAL_STATE_VARS];			/* system dynamic noise covariance matrix */
static double	R;												/* measurement noise variance */
static double	Inn,InnCov;										/* innovation and covariance of innovation */
static double	K[TOTAL_STATE_VARS];							/* Kalman gain matrix */
static double	S[TOTAL_STATE_VARS][TOTAL_STATE_VARS];			/* predicted covariance matrix */
static double	Supdated[TOTAL_STATE_VARS][TOTAL_STATE_VARS];	/* updated covariance matrix */
static double	temp[TOTAL_STATE_VARS][TOTAL_STATE_VARS];		/* used during solving matrix equations */

			  /* allocate memory for data buffer */
if (sIBI == -1)
  {
  DataBuffer=(unsigned short *)calloc(MAX_DATA,sizeof(unsigned short));
  (*TotalData)=0;
  if (*TotalMinutes == 0)
	{		  /* initialize log-normal population statistics */
	*MeanRSA=0.0;
	*StdDevRSA=0.0;
	TotalRSA=0;
	for (i=0; i<1000; i++)
	  NormHist[i]=0;
	}
  else		  /* if TotalMinutes comes in >0 on initialization, assume using calibrated stats */
	TotalRSA=(*TotalMinutes)*LocalSamplingRate*60;

			  /* intialize Kalman filter matrices for smoothing arousal */
  		/* initialize state transition matrix */
  phi[0][0]=1.0; phi[0][1]=DELTA_T;
  phi[1][0]=0.0; phi[1][1]=1.0;
		/* initialize measurement matrix */
  M[0]=1.0; M[1]=0.0;		/* measuring position only */
		/* initialize state matrix */
  Xtt[0]=0.0; Xtt[1]=0.0;		/* don't know where it is */
  Xt1t[0]=0.0; Xt1t[1]=0.0;	/* no prediction yet either */
		/* intialize system dynamic model noise covariance matrix */
  Qt[0][0]=Qt[0][1]=Qt[1][0]=0.0;
  Qt[1][1]=SQR(DELTA_T*MAX_ACCEL/ArousalSmooth);	/* sigmaU squared */
		/* initialize predictor covariance matrix */
  for (i=0; i<TOTAL_STATE_VARS; i++)
    for (j=0; j<TOTAL_STATE_VARS; j++)
      S[i][j]=Qt[i][j];
		/* intialize measurement error covariance */
  R=SQR(MEAS_NOISE);

  return(LocalWindowSize);
  }

			  /* calling with an sIBI=-2 shuts the local memory down */
if (sIBI == -2)
  {
  free(DataBuffer);
  return(0);
  }

			  /* initialize all returned values to zero */
for (i=0; i<TOTAL_FREQS; i++)
  Power[i]=0.0;
(*CurrentMaxPower)=0.0;
(*CurrentMaxPowerCPM)=0;
(*Arousal)=0.0;

			  /* buffer sIBI data until there are enough to analyze */
if (*TotalData >= LocalWindowSize)
  {			  /* push out oldest value from beginning of list, add new value to end of list */
  for (i=0; i<(*TotalData)-1; i++)
    DataBuffer[i]=DataBuffer[i+1];
  DataBuffer[(*TotalData)-1]=sIBI;
  }
else
  {
  DataBuffer[*TotalData]=sIBI;
  (*TotalData)++;
  }

			  /* when there is enough data, analyze the power spectrum */
if ((*TotalData) >= LocalWindowSize)
  {
  PowerAnalyze(DataBuffer,Power,LocalSamplingRate,LocalWindowSize);
  }
else
  {
  return(LocalWindowSize-(*TotalData));
  }


			  /* detect max power in 9-30 CPM range */
*CurrentMaxPowerCPM=10; // 9;
for (i=10; i<TOTAL_FREQS; i++)
  if (Power[i] > Power[*CurrentMaxPowerCPM])
    *CurrentMaxPowerCPM=i;
*CurrentMaxPower=Power[*CurrentMaxPowerCPM];

			  /* take log of max power, and add it into histogram */
LogPower=log(*CurrentMaxPower);
hist_bin_width=(double)(NormDistMaxValue-NormDistMinValue)/1000.0;
bin=(int)((LogPower-(double)NormDistMinValue)/hist_bin_width);
if (bin >= 0  &&  bin < 1000)
  NormHist[bin]++;

			  /* update RSA statistics over all time */
if (CalibrationLocked == 0)
  {
  SumRSA=(SQR(*StdDevRSA)+SQR(*MeanRSA))*(double)TotalRSA;
  *MeanRSA=(*MeanRSA*(double)TotalRSA+LogPower)/((double)TotalRSA+1.0);
  if (TotalRSA == 0)
    *StdDevRSA=0.0;
  else
    *StdDevRSA=sqrt((SumRSA+SQR(LogPower))/((double)TotalRSA+1.0)-SQR(*MeanRSA));
  *TotalMinutes=TotalRSA/(60*LocalSamplingRate);
  TotalRSA++;
  }

			  /* calculate Z-score of RSA for current time */
if (*StdDevRSA < 0.001) /* should only happen on first reading, for which stddev is zero */
  ZScore=0.0;
else
  ZScore=(LogPower-*MeanRSA)/(*StdDevRSA);

			  /* filter arousal result (0.0-ZScore is the data) */
	/* innovation (diff between last prediction and new measurement) */
Inn=(0.0-ZScore)-Xt1t[0];	  /* 0.0-ZScore so that higher value = larger arousal */
	/* covariance of innovation  MSM' + R  (M' = M transpose) */
InnCov=(M[0]*S[0][0]+M[1]*S[1][0])*M[0]+(M[0]*S[0][1]+M[1]*S[1][1])*M[1]+R;
	/* Kalman gain (weight) equation  K=S*M'*inv(InnCov) */
K[0]=(S[0][0]*M[0]+S[0][1]*M[1])*(1.0/InnCov);
K[1]=(S[1][0]*M[0]+S[1][1]*M[1])*(1.0/InnCov);
	/* filtering equation (update state estimate)  Xtt=Xt1t+K*Inn */
Xtt[0]=Xt1t[0]+K[0]*Inn;
Xtt[1]=Xt1t[1]+K[1]*Inn;
	/* update covariance of prediction */
Supdated[0][0]=(1.0-K[0])*S[0][0]+(0.0-0.0)*S[1][0];
Supdated[0][1]=(1.0-K[0])*S[0][1]+(0.0-0.0)*S[1][1];
Supdated[1][0]=(0.0-K[1])*S[0][0]+(1.0-0.0)*S[1][0];
Supdated[1][1]=(0.0-K[1])*S[0][1]+(1.0-0.0)*S[1][1];
	/* predictor equation  Xt1t=phi*Xtt */
Xt1t[0]=phi[0][0]*Xtt[0]+phi[0][1]*Xtt[1];
Xt1t[1]=phi[1][0]*Xtt[0]+phi[1][1]*Xtt[1];
	/* predict covariance  S=phi*Supdated*phi'+Q */
temp[0][0]=phi[0][0]*Supdated[0][0]+phi[0][1]*Supdated[1][0];
temp[0][1]=phi[0][0]*Supdated[0][1]+phi[0][1]*Supdated[1][1];
temp[1][0]=phi[1][0]*Supdated[0][0]+phi[1][1]*Supdated[1][0];
temp[1][1]=phi[1][0]*Supdated[0][1]+phi[1][1]*Supdated[1][1];
S[0][0]=(temp[0][0]*phi[0][0]+temp[0][1]*phi[0][1])+Qt[0][0];
S[0][1]=(temp[0][0]*phi[1][0]+temp[0][1]*phi[1][1])+Qt[0][1];
S[1][0]=(temp[1][0]*phi[0][0]+temp[1][1]*phi[0][1])+Qt[1][0];
S[1][1]=(temp[1][0]*phi[1][0]+temp[1][1]*phi[1][1])+Qt[1][1];

			  /* filtered arousal is output of filter (state is Xtt[0],Xtt[1]) */
*Arousal=Xtt[0];

return(0);  /* arousal is "live" */

}





void PowerAnalyze(short int *Data,double *Power,int SetSamplingRate,int SetWindowSize)

{
double	*MeanCenteredData,*HammingCenteredData,*HammingWeight;
double	Mean,TimePerDataPoint;
int		PointsInWinSec;
complex	*FourierTransformed;
int		i,j;
int		TotalDataToAnalyze,CPM;
double	PowerCheck[30];

	/* make analysis statistics (time variables) */
TimePerDataPoint=1.0/(double)SetSamplingRate;
PointsInWinSec=(int)((double)SetWindowSize/TimePerDataPoint+0.5);
TotalDataToAnalyze=SetWindowSize;

	/* mean-center the windowed data at zero */
MeanCenteredData=(double *)calloc(TotalDataToAnalyze,sizeof(double));
HammingCenteredData=(double *)calloc(TotalDataToAnalyze,sizeof(double));
Mean=0.0;
for (i=0; i<=TotalDataToAnalyze-1; i++)
  Mean+=(double)Data[i];
Mean/=(double)TotalDataToAnalyze;
for (i=0; i<TotalDataToAnalyze; i++)
  MeanCenteredData[i]=(double)Data[i]-Mean;

	/* apply Hamming window to entire window of data */
for (i=0; i<TotalDataToAnalyze; i++)
  HammingCenteredData[i]=MeanCenteredData[i];
HammingWeight=(double *)calloc(TotalDataToAnalyze/2,sizeof(double));
for (i=0; i<TotalDataToAnalyze/2; i++)
  HammingWeight[i]=0.54+0.46*cos((2.0*M_PI*(double)i)/
                ((double)TotalDataToAnalyze-1.0));
for (i=0,j=TotalDataToAnalyze/2-1; i<TotalDataToAnalyze/2; i++,j--)
  HammingCenteredData[i]=HammingWeight[j]*HammingCenteredData[i];
for (i=TotalDataToAnalyze/2,j=0; i<TotalDataToAnalyze; i++,j++)
  HammingCenteredData[i]=HammingWeight[j]*HammingCenteredData[i];
free(HammingWeight);

	/* FFT (forward) */
FourierTransformed=(complex *)calloc(TotalDataToAnalyze,sizeof(complex));
for (i=0; i<TotalDataToAnalyze; i++)
  {
  FourierTransformed[i].real=HammingCenteredData[i];
  FourierTransformed[i].imag=0.0;
  }
fast_fourier_transform(FourierTransformed,TotalDataToAnalyze);

for (i=0; i<TOTAL_FREQS; i++)
  Power[i]=0.0;

for (i=0; i<TOTAL_FREQS; i++)
  PowerCheck[i]=(SQR(FourierTransformed[i].real)+
		SQR(FourierTransformed[i].imag));

for (i=0; i<TotalDataToAnalyze; i++)
  {
//  CPM=(int)((double)i*1.0/(double)TotalDataToAnalyze*(double)SetSamplingRate*60.0);
  CPM=i;
  if (CPM >= TOTAL_FREQS)
	break;	/* don't care about the rest */
//  Power[CPM]+=(SQR(FourierTransformed[i].real)+SQR(FourierTransformed[i].imag));
  Power[CPM]=(SQR(FourierTransformed[i].real)+SQR(FourierTransformed[i].imag));
  }

free(MeanCenteredData);
free(HammingCenteredData);
free(FourierTransformed);
}









/*************************************************************************/
/* Function fast_fourier_transform will perform the single dimensional   */
/* fourier transform on the input array of complex points, the length of */
/* which (power of 2) is also passed in. The set of fourier coefficients */
/* from zero to N - 1 is returned in the same array, where each complex  */
/* value represents the phase and magnitude of the given frequency term. */
/* The algorithm uses the successive doubling technique of computing two */
/* point transforms and building upon them until the final transform is  */
/* accomplished. The algorithm is taken from Digital Image Processing by */
/* Rafael C. Gonzalez and Paul Wintz.                                    */
/*************************************************************************/

void fast_fourier_transform (freq,n)

complex *freq;              /* on input is complex points */
                            /* on output is coefficients, 0 to n - 1 */
int n;                      /* number of points in array */

{
 complex u,w,t;                         /* various complex numbers */
 double temp,n_inverse;                 /* temporary variables */
 int nv2,nm1,i,ip,j,j1,k,l,ln,le,le1;   /* limits,counters and indices */

 ln=0;
 j=n;
 while (j > 1)
   {
   if (j % 2 != 0)
     {
     printf("FFT must have power-of-two #points\n");
     exit(0);
     }
   j/=2;
   ln++;
   }
 //ln = (int)(log((double)n)/log(2.0)); /* find power of 2 of # of points */
 nv2 = n / 2;                       /* set pointer to middle of array */
 nm1 = n - 1;                       /* set pointer to next to last element */
 j = 1;                             /* set pointer at front of array */

 /* This loop is responsible for reordering the input array elements */
 /* so that the successive doubling method will work. Array elements */
 /* are swapped with their bit-wise opposites.                       */

 for (i = 0; i < nm1; i++)          /* make all swaps */
 {
  j1 = j - 1;                       /* temporary pointer */

  if (i < j1)                       /* check if bit-wise opposite found */
  {
   temp = freq[j1].real;            /* swap real parts */
   freq[j1].real = freq[i].real;
   freq[i].real = temp;
   temp = freq[j1].imag;            /* swap imaginary parts */
   freq[j1].imag = freq[i].imag;
   freq[i].imag = temp;
  }

  k = nv2;                          /* reset pointer to middle of array */

  /* This loop will find the next position at which to look for the */
  /* bit-wise opposite which is not itself.                         */

  while (k < j)                     /* loop until desired spot found */
  {
   j = j - k;                       /* reset indexes to get j pointing at */
   k = k / 2;                       /* next start point for the exchange */
  }

  j = j + k;                        /* increment pointer to bit-wise opposite */
 }

 /* This loop will perform the successive two point transforms on */
 /* corresponding elements in the array. Each pass the spread of  */
 /* the transforms increases, going from many two point transforms*/
 /* to the final two point transform.                             */

 for (l = 1; l <= ln; l++)                /* make ln passes */
 {
  le = (int)(pow(2.0,(double)l));         /* set spread of transforms to do */
  le1 = le / 2;                           /* set index of w factor */

  u.real = 1.0;                           /* set u = 1 + j0 */
  u.imag = 0.0;

  w.real = cos(M_PI / (double)le1);       /* set w = e^(-jpi/le1) */
  w.imag = - sin(M_PI / (double)le1);

  /* This loop performs the two point transforms of this pass */

  for (j = 1; j <= le1; j++)             /* do le1 transforms this pass */
  {
   for (i = j - 1; i < n; i = i + le)    /* do transforms le apart in array */
   {
    ip = i + le1;                        /* set index to partner of transform */

    /* calculate odd term times exponential factor */

    t.real = u.real * freq[ip].real - u.imag * freq[ip].imag;
    t.imag = u.imag * freq[ip].real + u.real * freq[ip].imag;

    freq[ip].real = freq[i].real - t.real;  /* calculate term F(u + M) */
    freq[ip].imag = freq[i].imag - t.imag;

    freq[i].real = freq[i].real + t.real;   /* calculate term F(u) */
    freq[i].imag = freq[i].imag + t.imag;
   }

   /* update exponential factor of transform terms for next pass */

   t.real = u.real * w.real - u.imag * w.imag;
   t.imag = u.imag * w.real + u.real * w.imag;
   u.real = t.real;
   u.imag = t.imag;
  }
 }

 n_inverse = 1.0 / (double)n;    /* calculate 1 / n */

 /* This loop will divide all coefficients by the number of coefficients */

 for (i = 0; i < n; i++)           /* do each coefficient */
 {
  freq[i].real = freq[i].real * n_inverse;
  freq[i].imag = freq[i].imag * n_inverse;
 }
}



